from __future__ import annotations

import logging
from functools import lru_cache
from typing import List, Sequence

from ldap3.utils.conv import escape_filter_chars

from ucsschool.apis.utils import LDAPAccess, LDAPCredentials, LDAPSettings, get_logger

from .models import RawUser
from .utils import ldap_credentials, ldap_settings

logger: logging.Logger = get_logger()
CONCATINATED_USERS_SEARCH_LIMIT = 50


@lru_cache(maxsize=1)
def ldap_access(settings: LDAPSettings = None, credentials: LDAPCredentials = None) -> LDAPAccess:
    """Cached LDAP access object"""
    return LDAPAccess(settings or ldap_settings(), credentials or ldap_credentials())


async def get_users(usernames: Sequence[str]) -> List[RawUser]:
    """Retrieve data of multiple users."""
    if not usernames:
        return []
    if len(usernames) > CONCATINATED_USERS_SEARCH_LIMIT:
        # recursion
        res = await get_users(usernames[:CONCATINATED_USERS_SEARCH_LIMIT])
        res.extend(await get_users(usernames[CONCATINATED_USERS_SEARCH_LIMIT:]))
        return res
    ldap = ldap_access()
    filter_uids = [f"(uid={escape_filter_chars(uid)})" for uid in usernames]
    filter_s = f"(&(univentionObjectType=users/user)(|{''.join(filter_uids)}))"
    res = []
    for result in await ldap.search(filter_s, attributes=["*"]):
        ucsschool_roles = result.ucsschoolRole.value
        if isinstance(ucsschool_roles, str):
            ucsschool_roles = [ucsschool_roles]
        schools = result.ucsschoolSchool.value
        if isinstance(schools, str):
            schools = [schools]
        res.append(
            RawUser(
                dn=result.entry_dn,
                firstname=result.givenName.value,
                lastname=result.sn.value,
                record_uid=result.ucsschoolRecordUID.value,
                schools=schools,
                source_uid=result.ucsschoolSourceUID.value,
                synonyms={
                    k: v[0]
                    for k, v in result.entry_attributes_as_dict.items()
                    if k.startswith("idBrokerPseudonym00")
                },
                ucsschool_roles=ucsschool_roles,
                username=result.uid.value,
            )
        )
    return res
