import datetime
from typing import Dict, List, Sequence
from unittest.mock import AsyncMock

import pytest
from fastapi.testclient import TestClient

import id_broker_common.kelvin
from id_broker_common.kelvin import kelvin_session
from id_broker_common.models import RawUser
from self_disclosure_plugin import dependencies
from self_disclosure_plugin.models import ClassMembers, Student, Teacher
from self_disclosure_plugin.routes.v1 import users as users_module
from self_disclosure_plugin.routes.v1.groups import router as groups_router
from self_disclosure_plugin.routes.v1.users import router as users_router
from ucsschool.apis.models import OAuth2Token
from ucsschool.apis.opa import opa_instance
from ucsschool.kelvin.client import SchoolClass

from ..utils import Dictable, DictObject, get_preconfigured_session_mock


@pytest.fixture
def client(app, dependency_overrides, fake_opa, async_dict):
    def _func(router, sub: str, client_id: str):
        app.include_router(router)
        overrides = {
            opa_instance: fake_opa,
            dependencies.token_dep: lambda: OAuth2Token(
                iss="",
                sub=sub,
                aud=[],
                exp=datetime.datetime.now() + datetime.timedelta(hours=3),
                client_id=client_id,
                iat=datetime.datetime.now() - datetime.timedelta(seconds=1),
                jti="",
            ),
            kelvin_session: get_preconfigured_session_mock(),
            dependencies.connecting_user_pseudonym_from_token: lambda: sub,
            dependencies.service_provider_from_token: lambda: client_id,
        }
        dependency_overrides(overrides)
        return TestClient(app)

    return _func


service_provider = "EinServiceProvider"
traeger1 = "EinTraeger"
service_provider_attribute = "idBrokerPseudonym0001"
unused_sp_attr = "idBrokerPseudonym0002"

student1 = DictObject(
    __name__=f"{traeger1}-student1",  # needed for parametrize
    udm_properties={service_provider_attribute: "stu1@pseudo", unused_sp_attr: None},
    name=f"{traeger1}-student1",
    roles=["student"],
    firstname="stu1",
    lastname="dent",
    school_classes={f"{traeger1}-demoschool": ["class1"]},
    school=f"{traeger1}-demoschool",
    schools=[f"{traeger1}-demoschool"],
    source_uid=traeger1,
    ucsschool_roles=[f"student:school:{traeger1}-demoschool"],
)
student2 = DictObject(
    __name__=f"{traeger1}-student2",  # needed for parametrize
    udm_properties={service_provider_attribute: "stu2@pseudo", unused_sp_attr: None},
    name=f"{traeger1}-student2",
    roles=["student"],
    firstname="stu2",
    lastname="dent",
    school_classes={f"{traeger1}-demoschool": ["class1"]},
    school=f"{traeger1}-demoschool",
    schools=[f"{traeger1}-demoschool"],
    source_uid=traeger1,
    ucsschool_roles=[f"student:school:{traeger1}-demoschool"],
)
student3 = DictObject(
    __name__=f"{traeger1}-student3",  # needed for parametrize
    udm_properties={service_provider_attribute: "stu3@pseudo", unused_sp_attr: None},
    name=f"{traeger1}-student3",
    roles=["student"],
    firstname="stu3",
    lastname="dent",
    school_classes={f"{traeger1}-demoschool": ["class2"]},
    school=f"{traeger1}-demoschool",
    schools=[f"{traeger1}-demoschool"],
    source_uid=traeger1,
    ucsschool_roles=[f"student:school:{traeger1}-demoschool"],
)
student4 = DictObject(
    __name__=f"{traeger1}-student4",  # needed for parametrize
    udm_properties={service_provider_attribute: None, unused_sp_attr: None},  # no pseudonym!
    name=f"{traeger1}-student4",
    roles=["student"],
    firstname="stu4",
    lastname="dent",
    school_classes={f"{traeger1}-demoschool": ["class2"]},
    school=f"{traeger1}-demoschool",
    schools=[f"{traeger1}-demoschool"],
    source_uid=traeger1,
    ucsschool_roles=[f"student:school:{traeger1}-demoschool"],
)
teacher1 = Dictable(
    __name__=f"{traeger1}-teacher1",  # needed for parametrize
    udm_properties={service_provider_attribute: "tea1@Pseudonym", unused_sp_attr: None},
    name=f"{traeger1}-teacher1",
    roles=["teacher"],
    firstname="tea1",
    lastname="cher",
    school_classes={f"{traeger1}-demoschool": ["class1"]},
    school=f"{traeger1}-demoschool",
    schools=[f"{traeger1}-demoschool"],
    source_uid=traeger1,
    ucsschool_roles=[f"teacher:school:{traeger1}-demoschool"],
)
class1 = Dictable(
    name="class1",
    users=[
        student1.name,
        student2.name,
        teacher1.name,
    ],
    school=f"{traeger1}-demoschool",
    udm_properties={
        service_provider_attribute: "class1@pseudo",
        unused_sp_attr: None,
        "ucsschoolRecordUID": "class1@record",
    },
)
class2 = Dictable(
    name="class2",
    users=[student3.name, student4.name],
    school=f"{traeger1}-demoschool",
    udm_properties={
        service_provider_attribute: "class2@pseudo",
        unused_sp_attr: None,
        "ucsschoolRecordUID": "class2@record",
    },
)
all_test_objects: Dict[str, DictObject] = {
    student1.name: student1,
    student2.name: student2,
    student3.name: student3,
    student4.name: student4,
    teacher1.name: teacher1,
    class1.name: class1,
    class2.name: class2,
}


async def get_kelvin_obj_mock(resource, name, school=None):
    return all_test_objects[name]


def user_to_raw_user(user: DictObject) -> RawUser:
    return RawUser(
        dn="foo",
        firstname=user.firstname,
        lastname=user.lastname,
        record_uid="foo",
        schools=user.schools,
        source_uid=user.source_uid,
        synonyms={k: v for k, v in user.udm_properties.items() if k.startswith("idBrokerPseudonym00")},
        ucsschool_roles=user.ucsschool_roles,
        username=user.name,
    )


async def ldap_get_users_mock(user_names: Sequence[str]) -> List[RawUser]:
    return [user_to_raw_user(all_test_objects[name]) for name in user_names]


async def _get_obj_attr_by_pseudonym_mock(pseudonym, sp_attribute, object_class, name_attr):
    for test_obj in all_test_objects.values():
        if test_obj.udm_properties.get(sp_attribute) == pseudonym:
            if object_class == "ucsschoolGroup":
                return {"name": f"{test_obj.school}-{test_obj.name}", "school": test_obj.school}
            else:
                return {"name": test_obj.name, "school": test_obj.school}
    raise AssertionError(f"No test object with pseudonym {pseudonym!r} in attr {sp_attribute!r} found.")


def student_from_dict_object(obj: DictObject) -> Student:
    prefix = f"{obj.source_uid}-"
    return Student(
        user_id=obj.udm_properties[service_provider_attribute],
        username=obj.name.removeprefix(prefix),
        firstname=obj.firstname,
        lastname=obj.lastname,
    )


def teacher_from_dict_object(obj: DictObject) -> Teacher:
    prefix = f"{obj.source_uid}-"
    return Teacher(
        user_id=obj.udm_properties[service_provider_attribute],
        username=obj.name.removeprefix(prefix),
    )


def test_users_metadata(client, monkeypatch):
    monkeypatch.setattr(dependencies, "get_user", AsyncMock(return_value=student1))
    test_client = client(
        router=users_router,
        sub=student1.udm_properties[service_provider_attribute],
        client_id=service_provider,
    )
    response = test_client.get(f"/{student1.udm_properties[service_provider_attribute]}/metadata")
    assert response.status_code == 200, response.__dict__
    json_response = response.json()
    prefix = f"{student1.source_uid}-"
    expected_user = {
        "user_id": student1.udm_properties[service_provider_attribute],
        "username": student1.name.removeprefix(prefix),
        "firstname": student1.firstname,
        "lastname": student1.lastname,
        "type": student1.roles[0],
        "school_id": student1.school.removeprefix(prefix),
    }
    assert json_response == expected_user


def test_user_groups(client, monkeypatch):
    monkeypatch.setattr(dependencies, "get_user", AsyncMock(return_value=student1))
    monkeypatch.setattr(
        dependencies,
        "get_service_provider_attribute",
        AsyncMock(return_value=service_provider_attribute),
    )
    monkeypatch.setattr(users_module, "get_kelvin_obj", get_kelvin_obj_mock)
    monkeypatch.setattr(id_broker_common.kelvin, "get_kelvin_obj", get_kelvin_obj_mock)
    monkeypatch.setattr(users_module, "get_users", ldap_get_users_mock)
    monkeypatch.setattr(
        users_module,
        "get_service_provider_attribute",
        AsyncMock(return_value=service_provider_attribute),
    )
    test_client = client(
        router=users_router,
        sub=teacher1.udm_properties[service_provider_attribute],
        client_id=service_provider,
    )
    response = test_client.get(f"/{teacher1.udm_properties[service_provider_attribute]}/groups")
    assert response.status_code == 200, response.__dict__
    json_response = response.json()
    expected_groups = {
        "groups": [
            {
                "group_id": class1.udm_properties[service_provider_attribute],
                "name": class1.name,
                "student_count": 2,
            }
        ]
    }
    assert json_response == expected_groups


def test_group_users(client, monkeypatch):
    monkeypatch.setattr(
        dependencies,
        "get_service_provider_attribute",
        AsyncMock(return_value=service_provider_attribute),
    )
    monkeypatch.setattr(
        dependencies,
        "get_connected_user",
        AsyncMock(return_value=teacher1),
    )
    monkeypatch.setattr(dependencies, "get_user", AsyncMock(return_value=teacher1))
    monkeypatch.setattr(dependencies, "_get_obj_attr_by_pseudonym", _get_obj_attr_by_pseudonym_mock)
    monkeypatch.setattr(dependencies, "get_kelvin_obj", get_kelvin_obj_mock)
    monkeypatch.setattr(id_broker_common.kelvin, "get_kelvin_obj", get_kelvin_obj_mock)
    monkeypatch.setattr(dependencies, "get_users", ldap_get_users_mock)
    test_client = client(
        router=groups_router,
        sub=teacher1.udm_properties[service_provider_attribute],
        client_id=service_provider,
    )
    response = test_client.get(f"/{class1.udm_properties[service_provider_attribute]}/users")
    assert response.status_code == 200, response.__dict__
    json_response = response.json()
    expected_group = {
        "students": [
            {
                "user_id": student1.udm_properties[service_provider_attribute],
                "username": student1.name.removeprefix(f"{student1.source_uid}-"),
                "firstname": student1.firstname,
                "lastname": student1.lastname,
            },
            {
                "user_id": student2.udm_properties[service_provider_attribute],
                "username": student2.name.removeprefix(f"{student2.source_uid}-"),
                "firstname": student2.firstname,
                "lastname": student2.lastname,
            },
        ],
        "teachers": [
            {
                "user_id": teacher1.udm_properties[service_provider_attribute],
                "username": teacher1.name.removeprefix(f"{teacher1.source_uid}-"),
            }
        ],
    }
    assert json_response == expected_group


@pytest.mark.asyncio
async def test_extract_members(monkeypatch):
    monkeypatch.setattr(
        dependencies,
        "get_service_provider_attribute",
        AsyncMock(return_value=service_provider_attribute),
    )
    monkeypatch.setattr(dependencies, "get_kelvin_obj", get_kelvin_obj_mock)
    monkeypatch.setattr(dependencies, "get_users", ldap_get_users_mock)
    classmembers = await dependencies.extract_members(SchoolClass(**class1), service_provider)
    assert classmembers == ClassMembers(
        students=[student_from_dict_object(student1), student_from_dict_object(student2)],
        teachers=[teacher_from_dict_object(teacher1)],
    )


@pytest.mark.parametrize("user", [student1, teacher1])
def test_roles_in_school(user):
    roles = dependencies.roles_in_school(user.ucsschool_roles, user.school)
    assert roles == set(user.roles)


@pytest.mark.asyncio
async def test_extract_members_filtered(monkeypatch, caplog):
    """
    Test if extract filters users without service provider specific pseudonym and logs a warning for
    them.
    """
    monkeypatch.setattr(
        dependencies,
        "get_service_provider_attribute",
        AsyncMock(return_value=service_provider_attribute),
    )
    monkeypatch.setattr(dependencies, "get_kelvin_obj", get_kelvin_obj_mock)
    monkeypatch.setattr(dependencies, "get_users", ldap_get_users_mock)
    classmembers = await dependencies.extract_members(SchoolClass(**class2), service_provider)
    assert classmembers == ClassMembers(
        students=[student_from_dict_object(student3)],
        teachers=[],
    )
    assert (
        f"Ignoring user {student4.name!r} without pseudonym in {service_provider_attribute!r} (for "
        f"service provider {service_provider!r})."
    ) in caplog.text


def test_user_groups_filtered(client, monkeypatch, caplog):
    """Same test as `test_extract_members_filtered`, but using the HTTP resource."""
    monkeypatch.setattr(dependencies, "get_user", AsyncMock(return_value=student3))
    monkeypatch.setattr(
        dependencies,
        "get_service_provider_attribute",
        AsyncMock(return_value=service_provider_attribute),
    )
    monkeypatch.setattr(dependencies, "get_kelvin_obj", get_kelvin_obj_mock)
    monkeypatch.setattr(users_module, "get_kelvin_obj", get_kelvin_obj_mock)
    monkeypatch.setattr(users_module, "get_users", ldap_get_users_mock)
    monkeypatch.setattr(
        users_module,
        "get_service_provider_attribute",
        AsyncMock(return_value=service_provider_attribute),
    )
    test_client = client(
        router=users_router,
        sub=student3.udm_properties[service_provider_attribute],
        client_id=service_provider,
    )
    response = test_client.get(f"/{student3.udm_properties[service_provider_attribute]}/groups")
    assert response.status_code == 200, response.__dict__
    json_response = response.json()
    expected_groups = {
        "groups": [
            {
                "group_id": class2.udm_properties[service_provider_attribute],
                "name": class2.name,
                "student_count": 1,
            }
        ]
    }
    assert json_response == expected_groups


def test_group_users_filtered(client, monkeypatch, caplog):
    monkeypatch.setattr(
        dependencies,
        "get_service_provider_attribute",
        AsyncMock(return_value=service_provider_attribute),
    )
    monkeypatch.setattr(
        dependencies,
        "get_connected_user",
        AsyncMock(return_value=teacher1),
    )
    monkeypatch.setattr(dependencies, "get_user", AsyncMock(return_value=teacher1))
    monkeypatch.setattr(dependencies, "_get_obj_attr_by_pseudonym", _get_obj_attr_by_pseudonym_mock)
    monkeypatch.setattr(dependencies, "get_kelvin_obj", get_kelvin_obj_mock)
    monkeypatch.setattr(id_broker_common.kelvin, "get_kelvin_obj", get_kelvin_obj_mock)
    monkeypatch.setattr(dependencies, "get_users", ldap_get_users_mock)
    test_client = client(
        router=groups_router,
        sub=teacher1.udm_properties[service_provider_attribute],
        client_id=service_provider,
    )
    response = test_client.get(f"/{class2.udm_properties[service_provider_attribute]}/users")
    assert response.status_code == 200, response.__dict__
    json_response = response.json()
    expected_users = {
        "students": [
            {
                "user_id": student3.udm_properties[service_provider_attribute],
                "username": student3.name.removeprefix(f"{student3.source_uid}-"),
                "firstname": student3.firstname,
                "lastname": student3.lastname,
            },
        ],
        "teachers": [],
    }
    assert json_response == expected_users
    assert (
        f"Ignoring user {student4.name!r} without pseudonym in {service_provider_attribute!r} (for "
        f"service provider {service_provider!r})."
    ) in caplog.text
